@@ -2,13 +2,14 @@ import abc
 import collections
 import copy
 import inspect
+import itertools
 import json
 import re
 import warnings
 from datetime import datetime
 
 import elasticsearch_dsl as dsl
-import six
+
 from django.conf import settings
 from django.contrib import messages
 from django.forms.forms import Form
@@ -26,16 +27,14 @@ from django.views.generic.edit import CreateView, FormView
 from elasticsearch_dsl import Q
 from elasticsearch_dsl.utils import AttrList
 
+from .facets import TermsFacet, RangeFilter, TextFacet
 from .mapping import DEFAULT_ANALYZER
 from .signals import advanced_search_performed, search_complete
 from .templatetags.seeker import seeker_format
 
-
-
 seekerview_field_templates = {}
 
 
-
 class Column(object):
     """
     """
@@ -68,16 +67,19 @@ class Column(object):
                 self.template_obj = loader.get_template(self.template)
             else:
                 self.template_obj = self.view.get_field_template(self.field)
-        #Set the model_lower variable on Column to the lowercased name of the model on the mapping once view is set above
+        # Set the model_lower variable on Column to the lowercased name of the model on the mapping once view is set above
         try:
             self.model_lower = self.view.document._model
-        except Exception:
-            try:
-                self.model_lower = self.view.document.model.__name____lower__()
-                self.view.document._model = self.model_lower
-            except Exception:
-                self.model_lower = self.view.document.queryset().model.__name__.lower()
-                self.view.document._model = self.model_lower                
+        except AttributeError:
+            document = self.view.document
+            if hasattr(document, 'model'):
+                self.model_lower = document.model.__name__.lower()
+            elif hasattr(document, 'queryset'):
+                self.model_lower = document.queryset().model.__name__.lower()
+            else:
+                self.model_lower = ''
+            self.view.document._model = self.model_lower
+
         return self
 
     def header(self):
@@ -139,7 +141,7 @@ class Column(object):
                 if index_to_replace is not None:
                     modified_values[index_to_replace] = highlighted_value
             highlight = modified_values
-            
+
         if self.value_format:
             value = self.value_format(value)
             if highlight:
@@ -151,6 +153,7 @@ class Column(object):
             'value': value,
             'highlight': highlight,
             'model_lower': self.model_lower,
+            'doc_class_name': result.__class__.__name__.lower(),
             'view': self.view,
             'user': self.view.request.user,
             'query': self.view.get_keywords(self.view.request.GET),
@@ -220,7 +223,7 @@ class SeekerView(View):
     """
     A list of field names to exclude when generating columns.
     """
-    
+
     display = None
     """
     A list of field/column names to display by default.
@@ -282,7 +285,7 @@ class SeekerView(View):
     """
     The number of results to show per page.
     """
-    
+
     available_page_sizes = []
     """
     If set allows user to set options for page size to be changed, (must include the default page_size)
@@ -541,13 +544,14 @@ class SeekerView(View):
         search_templates = []
         if field_name in self.field_templates:
             search_templates.append(self.field_templates[field_name])
-        try:
-            search_templates.append('seeker/%s/%s.html' % (self.document.model.__name__.lower(), field_name))
-        except Exception:
-            search_templates.append('seeker/%s/%s.html' % (self.document.queryset().model.__name__.lower(), field_name))
+        if hasattr(self.document, 'model'):
+            search_templates.append('seeker/{}/{}.html'.format(self.document.model.__name__.lower(), field_name))
+        elif hasattr(self.document, 'queryset'):
+            search_templates.append('seeker/{}/{}.html'.format(self.document.queryset().model.__name__.lower(), field_name))
         for _cls in inspect.getmro(self.document):
             if issubclass(_cls, dsl.DocType):
-                search_templates.append('seeker/%s/%s.html' % (_cls._doc_type.name, field_name))
+                search_templates.append('seeker/{}/{}.html'.format(_cls.__name__.lower(), field_name))
+                search_templates.append('seeker/{}/{}.html'.format(_cls._doc_type.name, field_name))
         search_templates.append('seeker/column.html')
         template = loader.select_template(search_templates)
         existing_templates = list(set(self._field_templates.values()))
@@ -579,10 +583,31 @@ class SeekerView(View):
         sort = self.get_field_sort(field_name)
         highlight = self.get_field_highlight(field_name)
         header = self.custom_column_headers.get(field_name, None)
-        field_definition  = self.field_definitions.get(field_name)
+        field_definition = self.field_definitions.get(field_name)
         return Column(field_name, label=label, sort=sort, highlight=highlight, header=header, field_definition=field_definition)
 
-    def get_columns(self, display=None):
+    def sort_columns(self, columns, display=None):
+        if not display:
+            display = self.get_display()
+
+        if self.display_column_sort_order:
+            sort_order = self.get_sort_order(columns)
+            columns.sort(key=lambda c: sort_order.index(c.field))
+            return columns
+        else:
+            visible_columns = []
+            non_visible_columns = []
+            for c in columns:
+                if c.visible:
+                    visible_columns.append(c)
+                else:
+                    non_visible_columns.append(c)
+            visible_columns.sort(key=lambda  c: display.index(c.field))
+            non_visible_columns.sort(key=lambda c: c.label)
+
+            return visible_columns + non_visible_columns
+
+    def get_columns(self, display=None, sort=True):
         """
         Returns a list of :class:`seeker.Column` objects based on self.columns, converting any strings.
         """
@@ -596,7 +621,7 @@ class SeekerView(View):
         else:
             # Otherwise, go through and convert any strings to Columns.
             for c in self.columns:
-                if isinstance(c, six.string_types):
+                if isinstance(c, str):
                     if self.exclude and c in self.exclude:
                         continue
                     columns.append(self.make_column(c))
@@ -608,25 +633,13 @@ class SeekerView(View):
         if not display:
             display = self.get_display()
 
-        if self.display_column_sort_order:
-            for c in columns:
-                c.bind(self, c.field in display)
-            sort_order = self.get_sort_order(columns)
-            columns.sort(key=lambda c: sort_order.index(c.field))
-            return columns
-        else:
-            visible_columns = []
-            non_visible_columns=[]
-            for c in columns:
-                c.bind(self, c.field in display)
-                if c.visible:
-                    visible_columns.append(c)
-                else:
-                    non_visible_columns.append(c)
-            visible_columns.sort(key=lambda  c: display.index(c.field))
-            non_visible_columns.sort(key=lambda c: c.label)
+        for c in columns:
+            c.bind(self, c.field in display)
 
-            return visible_columns + non_visible_columns
+        if sort:
+            columns = self.sort_columns(columns, display)
+
+        return columns
 
     def get_sorted_display_list(self):
         return self.request.GET.getlist("so")
@@ -685,6 +698,12 @@ class SeekerView(View):
         from .models import SavedSearch
         return SavedSearch
 
+    def get_saved_searches(self):
+        if self.request.user and self.request.user.is_authenticated:
+            return self.request.user.seeker_searches.filter(url=self.request.path)
+        else:
+            return []
+
     def get_search_fields(self, mapping=None, prefix=''):
         if self.search:
             return self.search
@@ -693,8 +712,8 @@ class SeekerView(View):
             for field_name in mapping:
                 if mapping[field_name].to_dict().get('analyzer') == self.analyzer:
                     fields.append(prefix + field_name)
-                if hasattr(mapping[field_name], 'properties'):
-                    fields.extend(self.get_search_fields(mapping=mapping[field_name].properties, prefix=prefix + field_name + '.'))
+                if hasattr(mapping[field_name], '_mapping'):
+                    fields.extend(self.get_search_fields(mapping=mapping[field_name]._mapping, prefix=prefix + field_name + '.'))
             return fields
         else:
             return self.get_search_fields(mapping=self.document._doc_type.mapping)
@@ -725,9 +744,7 @@ class SeekerView(View):
                     facet.apply(s)
         return s
 
-    def sort_descriptor(self, sort):
-        if self.missing_sort is None or isinstance(sort, dict):
-            return sort
+    def apply_sort_descriptor(self, sort):
         desc = sort.startswith('-')
         field = sort.lstrip('-')
         missing = self.missing_sort
@@ -736,14 +753,33 @@ class SeekerView(View):
         elif missing == '_high':
             missing = '_first' if desc else '_last'
         return {
-            field: {
-                'order': 'desc' if desc else 'asc',
-                'missing': missing,
+                field: {
+                    'order': 'desc' if desc else 'asc',
+                    'missing': missing,
+                }
             }
-        }
+
+    def sort_descriptor(self, sort):
+        if self.missing_sort is None or isinstance(sort, dict):
+            return sort
+        if not isinstance(sort, list):
+            return self.apply_sort_descriptor(sort)
+        else:
+            sort_dict = {}
+            for s in sort:
+                sort_dict.update(self.apply_sort_descriptor(s))
+            return sort_dict
+
+    def is_initial(self):
+        """
+        Returns true if the request is in an initial search, false if not
+        Can be overridden in case of non initial get requests that aren't ajax (submitted form)
+        This will depend on the seeker form html and the info coming in on the get request
+        """
+        return not self.request.is_ajax()
 
     def apply_highlight(self, search, columns):
-        highlight_fields = self.highlight if isinstance(self.highlight, (list, tuple)) else [c.highlight for c in columns if c.visible and c.highlight]
+        highlight_fields = self.highlight if isinstance(self.highlight, (list, tuple)) else [c.highlight for c in columns if c.highlight]
         # NOTE: If the option to customize the tags (via pre_tags and post_tags) is added then the Column "render" function will need to be updated.
         search = search.highlight(*highlight_fields, number_of_fragments=self.number_of_fragments).highlight_options(encoder=self.highlight_encoder)
         return search
@@ -760,19 +796,16 @@ class SeekerView(View):
 
         # Figure out if this is a saved search, and grab the current user's saved searches.
         saved_search = None
-        if self.request.user and self.request.user.is_authenticated:
-            saved_search_pk = self.get_saved_search()
-            if saved_search_pk:
-                try:
-                    saved_search = self.request.user.seeker_searches.get(pk=saved_search_pk, url=self.request.path)
-                except SavedSearchModel.DoesNotExist:
-                    pass
-            saved_searches = self.request.user.seeker_searches.filter(url=self.request.path)
-        else:
-            saved_searches = []
+        saved_searches = self.get_saved_searches()
+        saved_search_pk = self.get_saved_search()
+        if saved_searches and saved_search_pk:
+            try:
+                saved_search = saved_searches.get(pk=saved_search_pk)
+            except SavedSearchModel.DoesNotExist:
+                pass
 
         keywords = self.get_keywords(self.request.GET)
-        facets = self.get_facet_data(self.request.GET, initial=self.initial_facets if not self.request.is_ajax() else None)
+        facets = self.get_facet_data(self.request.GET, initial=self.initial_facets if self.is_initial()  else None)
         search = self.get_search(keywords, facets)
         columns = self.get_columns()
 
@@ -780,7 +813,7 @@ class SeekerView(View):
             executed_search = search.execute()
             facets_selected_and_results = collections.OrderedDict()
             for facet in facets:
-                if self.request.GET.get(facet.field):
+                if facets[facet]:
                     stored_facet_data = facets[facet]
                     facets[facet] = []
                     facets_selected_and_results[facet] = (stored_facet_data, self.get_search(keywords, facets).execute())
@@ -812,15 +845,14 @@ class SeekerView(View):
         page_size = self.get_page_size()
         page = self.request.GET.get('p', '').strip()
         page = int(page) if page.isdigit() else 1
-        offset = (page - 1) * self.page_size
+        offset = (page - 1) * page_size
         results_count = search[0:0].execute().hits.total
-        if results_count < offset:
+        if results_count <= offset:
             page = 1
             offset = 0
 
         # Finally, grab the results.
         results = search.sort(*sort_fields)[offset:offset + page_size].execute()
-
         context_querystring = self.normalized_querystring(ignore=['p'])
         sort = sorts[0] if sorts else None
         context = {
@@ -949,7 +981,7 @@ class SeekerView(View):
             # A "sub" method that handles ajax save submissions (and returns JSON, not a redirect)
             if request.is_ajax() and self.use_save_form:
 
-                response_data = {} # All data must be able to flatten to JSON
+                response_data = {}  # All data must be able to flatten to JSON
 
                 # First we check if the user is attempting to overwrite an existing search
                 saved_searches = request.user.seeker_searches.filter(url=request.path)
@@ -970,6 +1002,7 @@ class SeekerView(View):
                     saved_search.querystring = qs
                     saved_search.url = request.path
                     saved_search.save()
+                    form.save_m2m()
                     messages.success(request, 'Successfully saved "%s".' % saved_search)
                     response_data['redirect_url'] = saved_search.get_absolute_url()
                 else:
@@ -1020,9 +1053,12 @@ class SeekerView(View):
 
 
 class AdvancedColumn(Column):
+
     def header(self, results=None):
-        cls = '%s_%s' % (self.view.document._doc_type.name, self.field.replace('.', '_'))
-        cls += ' %s_%s' % (self.model_lower, self.field.replace('.', '_'))
+        cls = '{}_{}'.format(self.view.document._doc_type.name, self.field.replace('.', '_'))
+        cls += ' {}_{}'.format(self.view.document.__name__.lower(), self.field.replace('.', '_'))
+        if self.model_lower:
+            cls += ' {}_{}'.format(self.model_lower, self.field.replace('.', '_'))
         if not self.sort:
             return mark_safe('<th class="%s">%s</th>' % (cls, self.header_html))
         current_sort = self.view.search_object['sort']
@@ -1036,7 +1072,7 @@ class AdvancedColumn(Column):
             data_sort = '{}{}'.format(d, self.field)
         else:
             data_sort = self.field
-            
+
         next_sort = 'descending' if sort == 'Ascending' else 'ascending'
         sr_label = (' <span class="sr-only">(%s)</span>' % sort) if sort else ''
 
@@ -1060,7 +1096,7 @@ class AdvancedColumn(Column):
         """
         max_length = 0
         for result in results.hits:
-            field_len = len(six.text_type(result[self.field])) if (self.field in result and result[self.field]) else 0
+            field_len = len(str(result[self.field])) if (self.field in result and result[self.field]) else 0
             max_length = max(max_length, field_len)
         return max_length
 
@@ -1106,7 +1142,7 @@ class AdvancedSeekerView(SeekerView):
     A list of available page sizes. The values must be integers.
     NOTE: These values are NOT enforced, they are simply used for rendering.
     """
-    
+
     boolean_translations = {
         'AND': 'must',
         'OR': 'should'
@@ -1146,9 +1182,15 @@ class AdvancedSeekerView(SeekerView):
     The number of seconds to allow any DSL search to execute before a timeout error is raised.
     """
 
+    always_display_highlighted_columns = False
+    """
+    If True, any column that includes a cell with highlighted text will be automaticaly displayed by the system.
+    """
+
     @abc.abstractproperty
     def save_search_url(self):
         pass
+
     """
     This property should return the url of the associated AdvancedSavedSearchView.
     It is set to abstract because it needs to be defined on a site by site basis. None is a valid value if saved searches are not being used.
@@ -1157,6 +1199,7 @@ class AdvancedSeekerView(SeekerView):
     @abc.abstractproperty
     def search_url(self):
         pass
+
     """
     This property should return the url of this seeker view.
     It is set to abstract because it needs to be defined on a site by site basis.
@@ -1185,6 +1228,16 @@ class AdvancedSeekerView(SeekerView):
     If True, aggregations will be executed in a separate DSL Search object.  This will allows sites to cache aggregation searches.
     """
 
+    default_column_class = AdvancedColumn
+    """
+    This is the default class that make_column will use when create seeker columns.
+    """
+
+    field_column_classes = {}
+    """
+    A dictionary of field_name -> AdvancedColumn class that allows you to easily customize a column class and still utilize the make_column method.
+    """
+
     def __init__(self):
         if vars(SeekerView).get('get_search_query_type') != getattr(self, 'get_search_query_type').__func__:
             warnings.warn(
@@ -1258,22 +1311,42 @@ class AdvancedSeekerView(SeekerView):
         highlight = self.get_field_highlight(field_name)
         header = self.custom_column_headers.get(field_name, None)
         val_format = self.value_formats.get(field_name, None)
-        field_definition  = self.field_definitions.get(field_name)
-        return AdvancedColumn(field_name, label=label, sort=sort, highlight=highlight, header=header, value_format=val_format, field_definition=field_definition)
+        field_definition = self.field_definitions.get(field_name)
+
+        column_class = self.field_column_classes.get(field_name, self.default_column_class)
+        return column_class(
+            field_name,
+            label=label,
+            sort=sort,
+            highlight=highlight,
+            header=header,
+            value_format=val_format,
+            field_definition=field_definition
+        )
+
+    def apply_sort_field(self, column_lookup, sort):
+        c = column_lookup.get(sort.lstrip('-'))
+        if c and c.sort:
+            return('-{}'.format(c.sort) if sort.startswith('-') else c.sort)
+        return sort
 
     def get_sort_field(self, columns, sort, display):
         """
         Returns the appropriate sort field for a given sort value.
         """
+
         # Make sure we sanitize the sort fields.
+        sort_fields = []
         column_lookup = { c.field: c for c in columns }
         # Order of precedence for sort is: parameter, the default from the view, and then the first displayed column (if any are displayed)
         sort = sort or self.sort or display[0] if len(display) else ''
         # Get the column based on the field name, and use it's "sort" field, if applicable.
-        c = column_lookup.get(sort.lstrip('-'))
-        if c and c.sort:
-            return '-{}'.format(c.sort) if sort.startswith('-') else c.sort
-        return sort
+        if not isinstance(sort, list):
+            return self.apply_sort_field(column_lookup, sort)
+        else:
+            for s in sort:
+                sort_fields.append(self.apply_sort_field(column_lookup, s))
+        return sort_fields
 
     def get_sorted_display_list(self):
         return self.search_object.get("displaySortOrder")
@@ -1281,11 +1354,12 @@ class AdvancedSeekerView(SeekerView):
     def get(self, request, *args, **kwargs):
         facets = self.get_facets()
         context = {
-            'can_save': self.can_save and self.request.user and self.request.user.is_authenticated(),
+            'can_save': self.can_save and self.request.user and self.request.user.is_authenticated,
             'facets': facets,
             'search_url': self.search_url,
             'save_search_url': self.save_search_url,
-            'selected_facets': self.initial_facets
+            'selected_facets': self.initial_facets,
+            'initial_search_object_query': self.initial_facet_query()
         }
 
         if self.extra_context:
@@ -1354,6 +1428,19 @@ class AdvancedSeekerView(SeekerView):
             search = self.get_search_query_type(search, keywords)
         return search
 
+    def display_highlighted_columns(self, columns, display, results):
+        highlight_results = [list(hit.meta.highlight.to_dict().keys()) for hit in results if hasattr(hit.meta, 'highlight')]
+        highlighted_columns = list(set(itertools.chain.from_iterable(highlight_results)))
+        self.search_object['system_added_display'] = []
+        for column in columns:
+            if not column.visible and column.field in highlighted_columns:
+                column.bind(self, True)
+                self.search_object['system_added_display'].append(column.field)
+
+        columns = self.sort_columns(columns, display + self.search_object['system_added_display'])
+
+        return columns
+
     def get_aggregated_results(self, query, facet_lookup):
         """This function gets the aggregation results for a query.  It does not get the search results"""
         # We build a whole new search because apply_aggregations somehow breaks the search object being "immutable"
@@ -1379,7 +1466,7 @@ class AdvancedSeekerView(SeekerView):
         }
 
         json_response = {
-            'filters': [facet.build_filter_dict(aggregation_results) for facet in facet_lookup.values()], # Relies on the default 'apply_aggregations' being applied.
+            'filters': [facet.build_filter_dict(aggregation_results) for facet in facet_lookup.values()],  # Relies on the default 'apply_aggregations' being applied.
             'search_object': self.search_object
         }
 
@@ -1387,6 +1474,15 @@ class AdvancedSeekerView(SeekerView):
 
         return JsonResponse(json_response)
 
+    def initial_facet_query(self):
+
+        initial_query = {'condition': self.initial_facets.get('condition', 'AND'), 'rules': []}
+        for facet in self.get_facets():
+            if facet.field in self.initial_facets:
+                if hasattr(facet, 'initialize') and self.initial_facets[facet.field]:
+                    initial_query['rules'].append(facet.initialize(self.initial_facets[facet.field]))
+        return json.dumps(initial_query)
+
     def render_results(self, export):
         facet_lookup, query, advanced_query, facets_searched = self._get_processing_data()
 
@@ -1416,7 +1512,7 @@ class AdvancedSeekerView(SeekerView):
         page, offset = self.calculate_page_and_offset(self.search_object['page'], page_size, search)
 
         display = self.get_display(self.search_object['display'], facets_searched)
-        columns = self.get_columns(display)
+        columns = self.get_columns(display, sort=not self.always_display_highlighted_columns)
         if export:
             return self.export(search, columns)
 
@@ -1430,17 +1526,24 @@ class AdvancedSeekerView(SeekerView):
         # Finally, grab the results.
         sort = self.get_sort_field(columns, self.search_object['sort'], display)
         if sort:
-            results = search.sort(self.sort_descriptor(sort))[offset:offset + page_size].params(request_timeout=self.search_timeout).execute()
+            if (self.missing_sort is None or isinstance(sort, dict)) and isinstance(sort, list):
+                results = search.sort(*self.sort_descriptor(sort))[offset:offset + page_size].params(request_timeout=self.search_timeout).execute()
+            else:
+                results = search.sort(self.sort_descriptor(sort))[offset:offset + page_size].params(request_timeout=self.search_timeout).execute()
         else:
             results = search[offset:offset + page_size].params(request_timeout=self.search_timeout).execute()
 
         if not self.separate_aggregation_search:
             aggregation_results = results
 
+        if self.always_display_highlighted_columns:
+            columns = self.display_highlighted_columns(columns, display, results)
+
         # TODO clean this up (may not need everything)
         context = {
             'columns': columns,
             'display_columns': [c for c in columns if c.visible],
+            'system_added_display': self.search_object.get('system_added_display', []),
             'facet_lookup': facet_lookup,
             'facets_searched': facets_searched,
             'footer_template': self.footer_template,
@@ -1461,7 +1564,7 @@ class AdvancedSeekerView(SeekerView):
         self.modify_results_context(context)
 
         json_response = {
-            'filters': [facet.build_filter_dict(aggregation_results) for facet in facet_lookup.values()], # Relies on the default 'apply_aggregations' being applied.
+            'filters': [facet.build_filter_dict(aggregation_results) for facet in facet_lookup.values()],  # Relies on the default 'apply_aggregations' being applied.
             'table_html': loader.render_to_string(self.results_template, context, request=self.request),
             'search_object': self.search_object
         }
@@ -1473,7 +1576,7 @@ class AdvancedSeekerView(SeekerView):
     def calculate_page_and_offset(self, page, page_size, search):
         offset = (page - 1) * page_size
         results_count = search[0:0].params(request_timeout=self.search_timeout).execute().hits.total
-        if results_count < offset:
+        if results_count <= offset:
             page = 1
             offset = 0
         return page, offset
@@ -1585,6 +1688,7 @@ class AdvancedSeekerView(SeekerView):
         A helper method called when ``_export`` is present in the http request. Returns a ``StreamingHttpResponse``
         that yields CSV data for all matching results.
         """
+
         def csv_escape(value):
             if isinstance(value, (list, tuple)):
                 value = '; '.join(force_text(v) for v in value)
@@ -1709,7 +1813,7 @@ class AdvancedSavedSearchView(View):
             if delete:
                 if saved_search:
                     saved_search.delete()
-                saved_search = None # 'saved_search' will still hold the python object of the saved search that was just deleted
+                saved_search = None  # 'saved_search' will still hold the python object of the saved search that was just deleted
             elif modify_default:
                 if modify_default == 'set':
                     saved_search.default = True
@@ -1731,6 +1835,7 @@ class AdvancedSavedSearchView(View):
                     saved_search = form.save(commit=False)
                     saved_search.user = request.user
                     saved_search.save()
+                    form.save_m2m()
 
                     # TODO - Look for way to prevent us from needing to rerun this query to update results
                     saved_searches = self.get_saved_searches(url, SavedSearchModel)
